@@ -38,6 +38,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>	/* for chmod() */
 #include <dirent.h>
+#include <fcntl.h>
 #include <unistd.h>
 #include <signal.h>
 
@@ -346,6 +347,8 @@ static inline void do_chown (uid_t uid)
 /* Update a single file in an existing database. */
 static int update_one_file (const char *manpath, const char *filename)
 {
+	MYDBM_FILE dbf;
+
 	dbf = MYDBM_RWOPEN (database);
 	if (dbf) {
 		struct mandata info;
@@ -354,16 +357,15 @@ static int update_one_file (const char *manpath, const char *filename)
 		memset (&info, 0, sizeof (struct mandata));
 		manpage = filename_info (filename, &info, "");
 		if (info.name) {
-			dbdelete (info.name, &info);
-			purge_pointers (info.name);
+			dbdelete (dbf, info.name, &info);
+			purge_pointers (dbf, info.name);
 			free (info.name);
 		}
 		free (manpage);
 
-		test_manfile (filename, manpath);
+		test_manfile (dbf, filename, manpath);
 	}
 	MYDBM_CLOSE (dbf);
-	dbf = NULL;
 
 	return 1;
 }
@@ -438,12 +440,11 @@ static void cleanup (void *dummy ATTRIBUTE_UNUSED)
 	"#\thttp://www.brynosaurus.com/cachedir/\n"
 
 /* sort out the database names */
-static int mandb (const char *catpath, const char *manpath)
+static int mandb (const char *catpath, const char *manpath, int global_manpath)
 {
 	int ret, amount;
 	char *dbname;
-	char *cachedir_tag;
-	struct stat st;
+	int should_create;
 
 	dbname = mkdbname (catpath);
 	database = xasprintf ("%s/%d", catpath, getpid ());
@@ -452,34 +453,50 @@ static int mandb (const char *catpath, const char *manpath)
 		printf (_("Processing manual pages under %s...\n"), manpath);
 
 	if (!STREQ (catpath, manpath)) {
+		char *cachedir_tag;
+		int fd;
+
 		cachedir_tag = xasprintf ("%s/CACHEDIR.TAG", catpath);
-		if (stat (cachedir_tag, &st) == -1 && errno == ENOENT) {
+		fd = open (cachedir_tag, O_RDONLY);
+		if (fd < 0) {
 			FILE *cachedir_tag_file;
 
+			if (errno != ENOENT)
+				xremove (cachedir_tag);
 			cachedir_tag_file = fopen (cachedir_tag, "w");
 			if (cachedir_tag_file) {
 				fputs (CACHEDIR_TAG, cachedir_tag_file);
 				fclose (cachedir_tag_file);
+#ifdef SECURE_MAN_UID
+				if (global_manpath && euid == 0)
+					xchown (cachedir_tag,
+						man_owner->pw_uid, -1);
+#endif
+				xchmod (cachedir_tag, DBMODE);
 			}
-		}
+		} else
+			close (fd);
 		free (cachedir_tag);
 	}
 
+	should_create = (create || force_rescan || opt_test);
+
 #ifdef NDBM
 #  ifdef BERKELEY_DB
 	dbfile = xasprintf ("%s.db", dbname);
 	free (dbname);
 	tmpdbfile = xasprintf ("%s.db", database);
-	if (create || force_rescan || opt_test) {
+	if (!should_create) {
+		if (xcopy (dbfile, tmpdbfile) < 0)
+			should_create = 1;
+	}
+	if (should_create) {
 		xremove (tmpdbfile);
 		ret = create_db (manpath, catpath);
 		if (ret < 0)
 			return ret;
 		amount = ret;
 	} else {
-		ret = xcopy (dbfile, tmpdbfile);
-		if (ret < 0)
-			return ret;
 		ret = update_db_wrapper (manpath, catpath);
 		if (ret < 0)
 			return ret;
@@ -491,7 +508,12 @@ static int mandb (const char *catpath, const char *manpath)
 	free (dbname);
 	tmpdirfile = xasprintf ("%s.dir", database);
 	tmppagfile = xasprintf ("%s.pag", database);
-	if (create || force_rescan || opt_test) {
+	if (!should_create) {
+		if (xcopy (dirfile, tmpdirfile) < 0 ||
+		    xcopy (pagfile, tmppagfile) < 0)
+			should_create = 1;
+	}
+	if (should_create) {
 		xremove (tmpdirfile);
 		xremove (tmppagfile);
 		ret = create_db (manpath, catpath);
@@ -499,12 +521,6 @@ static int mandb (const char *catpath, const char *manpath)
 			return ret;
 		amount = ret;
 	} else {
-		ret = xcopy (dirfile, tmpdirfile);
-		if (ret < 0)
-			return ret;
-		ret = xcopy (pagfile, tmppagfile);
-		if (ret < 0)
-			return ret;
 		ret = update_db_wrapper (manpath, catpath);
 		if (ret < 0)
 			return ret;
@@ -514,16 +530,17 @@ static int mandb (const char *catpath, const char *manpath)
 #else /* !NDBM */
 	xfile = dbname; /* steal memory */
 	xtmpfile = database;
-	if (create || force_rescan || opt_test) {
+	if (!should_create) {
+		if (xcopy (xfile, xtmpfile) < 0)
+			should_create = 1;
+	}
+	if (should_create) {
 		xremove (xtmpfile);
 		ret = create_db (manpath, catpath);
 		if (ret < 0)
 			return ret;
 		amount = ret;
 	} else {
-		ret = xcopy (xfile, xtmpfile);
-		if (ret < 0)
-			return ret;
 		ret = update_db_wrapper (manpath, catpath);
 		if (ret < 0)
 			return ret;
@@ -583,7 +600,7 @@ static int process_manpath (const char *manpath, int global_manpath,
 	push_cleanup (cleanup, NULL, 0);
 	push_cleanup (cleanup_sigsafe, NULL, 1);
 	if (run_mandb) {
-		int ret = mandb (catpath, manpath);
+		int ret = mandb (catpath, manpath, global_manpath);
 		if (ret < 0) {
 			amount = ret;
 			goto out;
